/**
 * 
 */
package com.btc.ep.dal.persist.impl;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.persistence.EntityManager;
import javax.persistence.FlushModeType;
import javax.persistence.NoResultException;
import javax.persistence.NonUniqueResultException;
import javax.persistence.Query;
import javax.persistence.TypedQuery;

import com.btc.ep.base.bl.ModelElement;
import com.btc.ep.base.transactions.TransactionContext;

/**
 * JPA Generic/Base Data Access Object
 * 
 * An instance of this class may not be set in another field of an object as in
 * a Repository. The instance must always get from the PersistenceManager
 * because when the profile will be deleted, a new PersistenceService would be
 * created in the PersistenceManager.
 * 
 * @author Jacek
 *
 */
public class PersistenceService {

	/* *****************************************************
	 * * Generics ****************************************************
	 */

	private TransactionContext transactionContext;

	/**
	 * create new Instance with entityManager
	 * 
	 * @param entityManager
	 *            The EntityMangager
	 */
	public PersistenceService(TransactionContext transactionContext) {
		this.transactionContext = transactionContext;
	}

	/**
	 * add/persist new object
	 * 
	 * @param obj
	 *            Entity to persist
	 * @param <T>
	 *            Type definition
	 */
	public <T extends ModelElement> void persist(T obj) {
		EntityManager em = getEntityManager();
		em.persist(obj);
	}

	/**
	 * merge changes
	 * 
	 * @param obj
	 *            Entity to merge
	 * @param <T>
	 *            Type definition
	 * @return the new and attached object
	 */
	public <T extends ModelElement> T update(T obj) {
		return getEntityManager().merge(obj);
	}

	/**
	 * refresh an entity
	 * 
	 * @param obj
	 *            Entity to refresh
	 * @param <T>
	 *            Type definition
	 */
	public <T extends ModelElement> void refresh(T obj) {
		EntityManager em = getEntityManager();
		em.refresh(obj);
	}

	/**
	 * merge changes Persisted objects will be updated, the unpersisted objects
	 * will be persist.
	 * 
	 * @param dtos
	 *            the list of dto objects
	 * @param <T>
	 *            Type definition
	 * 
	 * @return List of new and attached object
	 */
	public <T extends ModelElement> List<T> update(final List<T> dtos) {
		final List<T> list = new ArrayList<T>();

		for (T element : dtos) {
			list.add(update(element));
		}
		return list;
	}

	/**
	 * merge changes Persisted objects will be updated, the unpersisted objects
	 * will be persist.
	 * 
	 * @param dtos
	 *            the set of dto objects
	 * @param <T>
	 *            Type definition
	 * 
	 * @return Set of new and attached object
	 */
	public <T extends ModelElement> Set<T> update(final Set<T> dtos) {
		final Set<T> set = new HashSet<T>();

		for (T element : dtos) {
			set.add(update(element));
		}
		return set;
	}

	/**
	 * remove persisted object by id
	 * 
	 * @param clazz
	 *            Type of the object
	 * @param <T>
	 *            Type definition
	 * @param uid
	 *            UUID of the Entity to be removed
	 */
	public <T extends ModelElement> void remove(Class<T> clazz, String uid) {
		EntityManager em = getEntityManager();
		T obj = em.find(clazz, uid);
		em.remove(obj);
	}

	/**
	 * remove a list of persisted objects
	 * 
	 * @param dtos
	 *            the list of dto objects
	 * @param <T>
	 *            Type definition
	 */
	public <T extends ModelElement> void remove(final Set<T> dtos) {
		EntityManager em = getEntityManager();
		for (T element : dtos) {
			em.remove(element);
		}
	}

	/**
	 * remove given obj
	 * 
	 * @param obj
	 *            Object to be removed
	 * @param <T>
	 *            Type definition
	 * 
	 * @param obj
	 */
	public <T extends ModelElement> void remove(T obj) {
		EntityManager em = getEntityManager();
		em.remove(obj);
	}

	/**
	 * check if persistenceContext contains object
	 *
	 * @param obj
	 *            to be checked
	 * @param <T>
	 *            Type definition
	 * 
	 * @return true if persistenceContext contains obj
	 */
	public <T extends ModelElement> boolean contains(T obj) {
		EntityManager em = getEntityManager();
		boolean res = em.contains(obj);
		return res;
	}

	/**
	 * get/lookup object by id
	 * 
	 * @param clazz
	 *            Type of the results
	 * @param <T>
	 *            Type definition
	 * @param uid
	 *            UUID of the Entity
	 * @return The Entity Object
	 */
	public <T extends ModelElement> T get(Class<T> clazz, Object uid) {
		EntityManager em = getEntityManager();
		return em.find(clazz, uid);
	}

	/**
	 * find entity by primary key
	 *
	 * @param <T>
	 *            class type inherited from ModelElement
	 * @param clazz
	 *            type of the result
	 * @param uid
	 *            primary key
	 * @return ModelElement single Result
	 */
	public <T extends ModelElement> T find(Class<T> clazz, String uid) {

		Map<String, Object> map = new HashMap<>();
		map.put("objectdb.result-fetch", "LAZY");
		EntityManager em = getEntityManager();
		return em.find(clazz, uid, map);
	}

	/**
	 * find entity by primary key
	 *
	 * @param <T>
	 *            class type inherited from ModelElement
	 * @param clazz
	 *            type of the result
	 * @param uid
	 *            primary key
	 * @return ModelElement single Result
	 */
	public <T extends ModelElement> T getReference(Class<T> clazz, String uid) {

		Map<String, Object> map = new HashMap<>();
		map.put("objectdb.result-fetch", "LAZY");
		EntityManager em = getEntityManager();
		return em.getReference(clazz, uid);
	}

	/**
	 * get/lookup object by named query
	 * 
	 * @param query
	 *            query id of the named query
	 * @param clazz
	 *            Type of the results
	 * @param <T>
	 *            Type definition
	 * @param args
	 *            query arguments
	 * @return ModelElement JPQL single Result
	 * @throws NoResultException
	 *             if no ModelElement found
	 * @throws NonUniqueResultException
	 *             if more than one ModelElement found
	 */
	public <T extends ModelElement> T getModelElementByNamedQuery(String query,
			Class<T> clazz, Object... args) {
		EntityManager em = getEntityManager();
		TypedQuery<T> q = em.createNamedQuery(query, clazz);
		for (int i = 0; i < args.length; i++) {
			Object arg = args[i];
			if (arg instanceof String) {
				arg = asLike((String) arg);
			}
			q.setParameter(i + 1, arg);
		}
		return q.getSingleResult();
	}

	/**
	 * get/lookup objects by named query
	 * 
	 * @param query
	 *            query id of the named query
	 * @param clazz
	 *            Type of the results
	 * @param <T>
	 *            Type definition
	 * @param args
	 *            query arguments
	 * @return List<ModelElement> JPQL Result list
	 */
	public <T extends ModelElement> List<T> getModelElementsByNamedQuery(
			String query, Class<T> clazz, Object... args) {
		EntityManager em = getEntityManager();
		TypedQuery<T> q = em.createNamedQuery(query, clazz);
		q.setFlushMode(FlushModeType.AUTO);
		for (int i = 0; i < args.length; i++) {
			Object arg = args[i];
			if (arg instanceof String) {
				arg = asLike((String) arg);
			}
			q.setParameter(i + 1, arg);
		}
		return q.getResultList();
	}

	/**
	 * get/lookup ModelElement object by query
	 * 
	 * @param query
	 *            JPQL query
	 * @param clazz
	 *            Type of the results
	 * @param <T>
	 *            Type definition
	 * @param args
	 *            query arguments
	 * @return ModelElement JPQL single Result or otherwise NULL
	 * @throws NoResultException
	 *             if no ModelElement found
	 */
	public <T extends ModelElement> T getModelElementOrNullByQuery(
			String query, Class<T> clazz, Object... args) {
		EntityManager em = getEntityManager();
		TypedQuery<T> q = em.createQuery(query, clazz);
		for (int i = 0; i < args.length; i++) {
			Object arg = args[i];
			if (arg instanceof String) {
				arg = asLike((String) arg);
			}
			q.setParameter(i + 1, arg);
		}
		List<T> list = q.getResultList();

		switch (list.size()) {
		case 0:
			return null;
		case 1:
			return list.get(0);
		default:
			throw new NonUniqueResultException(query);
		}
	}

	/**
	 * get/lookup ModelElement object by query
	 * 
	 * @param query
	 *            JPQL query
	 * @param clazz
	 *            Type of the results
	 * @param <T>
	 *            Type definition
	 * @param args
	 *            query arguments
	 * @return ModelElement JPQL single Result
	 * @throws NoResultException
	 *             if no ModelElement found
	 * @throws NonUniqueResultException
	 *             if more than one ModelElement found
	 */
	public <T extends ModelElement> T getModelElementByQuery(String query,
			Class<T> clazz, Object... args) {
		EntityManager em = getEntityManager();
		TypedQuery<T> q = em.createQuery(query, clazz);
		for (int i = 0; i < args.length; i++) {
			Object arg = args[i];
			if (arg instanceof String) {
				arg = asLike((String) arg);
			}
			q.setParameter(i + 1, arg);
		}
		return q.getSingleResult();
	}

	/**
	 * get/lookup ModelElement objects by query
	 * 
	 * @param query
	 *            JPQL query
	 * @param clazz
	 *            Type of the results
	 * @param <T>
	 *            Type definition
	 * @param args
	 *            query arguments
	 * @return List<ModelElement> JPQL Result list
	 */
	public <T> List<T> getModelElementsByQuery(String query, Class<T> clazz,
			Object... args) {
		EntityManager em = getEntityManager();
		TypedQuery<T> q = em.createQuery(query, clazz);
		q.setHint("objectdb.result-fetch", "LAZY");
		for (int i = 0; i < args.length; i++) {
			Object arg = args[i];
			if (arg instanceof String) {
				arg = asLike((String) arg);
			}
			q.setParameter(i + 1, arg);
		}
		return q.getResultList();
	}

	/**
	 * get/lookup object by query
	 * 
	 * @param queryName
	 *            query id
	 * @param clazz
	 *            Type of the results
	 * @param <T>
	 *            Type definition
	 * @param args
	 *            query arguments
	 * @return JPQL single Result
	 * @throws NoResultException
	 *             if no ModelElement found
	 * @throws NonUniqueResultException
	 *             if more than one ModelElement found
	 */
	public <T> T getSingleResultByNamedQuery(String queryName, Class<T> clazz,
			Object... args) {
		EntityManager em = getEntityManager();
		TypedQuery<T> q = em.createNamedQuery(queryName, clazz);
		for (int i = 0; i < args.length; i++) {
			Object arg = args[i];
			if (arg instanceof String) {
				arg = asLike((String) arg);
			}
			q.setParameter(i + 1, arg);
		}
		return q.getSingleResult();
	}

	/**
	 * get/lookup object by query
	 * 
	 * @param query
	 *            query
	 * @param clazz
	 *            Type of the results
	 * @param <T>
	 *            Type definition
	 * @param args
	 *            query arguments
	 * @return JPQL single Result
	 * @throws NoResultException
	 *             if no ModelElement found
	 * @throws NonUniqueResultException
	 *             if more than one ModelElement found
	 */
	public <T> T getSingleResultByQuery(String query, Class<T> clazz,
			Object... args) {
		EntityManager em = getEntityManager();
		TypedQuery<T> q = em.createQuery(query, clazz);
		for (int i = 0; i < args.length; i++) {
			Object arg = args[i];
			if (arg instanceof String) {
				arg = asLike((String) arg);
			}
			q.setParameter(i + 1, arg);
		}
		return q.getSingleResult();
	}

	/**
	 * get/lookup object by query
	 * 
	 * @param query
	 *            query
	 * @param clazz
	 *            Type of the results
	 * @param <T>
	 *            Type definition
	 * @param args
	 *            query arguments
	 * @return JPQL single Result
	 * @throws NoResultException
	 *             if no ModelElement found
	 * @throws NonUniqueResultException
	 *             if more than one ModelElement found
	 */
	public <T> T getSpecialSingleResultByQuery(String query, Class<T> clazz,
			Object... args) {
		EntityManager em = getEntityManager();
		TypedQuery<T> q = em.createQuery(query, clazz);

		// EntityGraph<T> entityGraph = em.createEntityGraph(clazz);
		// entityGraph.addAttributeNodes(graphs);
		// q.setHint("javax.persistence.fetchgraph", entityGraph);
		q.setHint("objectdb.result-fetch", "LAZY");
		for (int i = 0; i < args.length; i++) {
			Object arg = args[i];
			if (arg instanceof String) {
				arg = asLike((String) arg);
			}
			q.setParameter(i + 1, arg);
		}
		return q.getSingleResult();
	}

	/**
	 * get/lookup objects by query
	 * 
	 * @param query
	 *            query id
	 * @param clazz
	 *            Type of the results
	 * @param <T>
	 *            Type definition
	 * @param args
	 *            query arguments
	 * @return JPQL Result list
	 */
	public <T> List<T> getResultListByQuery(String query, Class<T> clazz,
			Object... args) {
		EntityManager em = getEntityManager();
		TypedQuery<T> q = em.createQuery(query, clazz);
		for (int i = 0; i < args.length; i++) {
			Object arg = args[i];
			if (arg instanceof String) {
				arg = asLike((String) arg);
			}
			q.setParameter(i + 1, arg);
		}
		return q.getResultList();
	}

	/**
	 * execute an update by query
	 * 
	 * @param queryName
	 *            query id
	 * @param args
	 *            query arguments
	 */
	public void updateByNamedQuery(String queryName, Object... args) {

		EntityManager em = getEntityManager();
		Query query = em.createNamedQuery(queryName);
		query.executeUpdate();
	}

	/**
	 * execute an update query
	 * 
	 * @param query
	 *            query
	 * @param args
	 *            query arguments
	 */
	public void updateQuery(String query, Object... args) {

		EntityManager em = getEntityManager();
		Query q = em.createQuery(query);
		for (int i = 0; i < args.length; i++) {
			Object arg = args[i];
			if (arg instanceof String) {
				arg = asLike((String) arg);
			}
			q.setParameter(i + 1, arg);
		}
		q.executeUpdate();
	}


	/* *****************************************************
	 * * Helper methods ****************************************************
	 */

	/**
	 * access to entityManager
	 *
	 * @return EntityManager
	 */
	private EntityManager getEntityManager() {
		return transactionContext.getEntityManager();
	}

	/**
	 * map LIKE wildcards
	 * 
	 * @param name
	 *            input String
	 * @return replaced resul
	 */
	private String asLike(String name) {
		String al = name.replace('*', '%');
		return al;
	}

}
